import java.awt.BorderLayout;
import java.awt.Color;
import java.awt.Desktop;
import java.awt.GraphicsEnvironment;
import java.awt.SystemColor;
import java.awt.Toolkit;
import java.awt.datatransfer.Clipboard;
import java.awt.datatransfer.ClipboardOwner;
import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.StringSelection;
import java.awt.datatransfer.Transferable;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.text.NumberFormat;
import java.util.EventObject;
import java.util.Objects;
import java.util.Random;

import javax.imageio.ImageIO;
import javax.swing.DefaultListModel;
import javax.swing.JButton;
import javax.swing.JDesktopPane;
import javax.swing.JFrame;
import javax.swing.JInternalFrame;
import javax.swing.JList;
import javax.swing.JMenu;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JOptionPane;
import javax.swing.JPanel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.SwingUtilities;
import javax.swing.UIManager;
import javax.swing.WindowConstants;
import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.TransformerException;

import org.w3c.dom.Document;
import org.xml.sax.SAXException;

import com.mindfusion.common.DateTime;
import com.mindfusion.common.StringUtilities;
import com.mindfusion.drawing.Colors;
import com.mindfusion.drawing.DashStyle;
import com.mindfusion.drawing.SolidBrush;
import com.mindfusion.spreadsheet.Cell;
import com.mindfusion.spreadsheet.CellAnnotation;
import com.mindfusion.spreadsheet.CellChangedEvent;
import com.mindfusion.spreadsheet.CellData;
import com.mindfusion.spreadsheet.CellEvent;
import com.mindfusion.spreadsheet.CellMouseEvent;
import com.mindfusion.spreadsheet.CellRange;
import com.mindfusion.spreadsheet.CellValidationEvent;
import com.mindfusion.spreadsheet.ColumnMouseEvent;
import com.mindfusion.spreadsheet.CsvImporter;
import com.mindfusion.spreadsheet.DefaultNumericFormats;
import com.mindfusion.spreadsheet.FormattedText;
import com.mindfusion.spreadsheet.HorizontalAlignment;
import com.mindfusion.spreadsheet.HyperlinkType;
import com.mindfusion.spreadsheet.IStyle;
import com.mindfusion.spreadsheet.IValidation;
import com.mindfusion.spreadsheet.InplaceEditEvent;
import com.mindfusion.spreadsheet.InplaceEditValidationEvent;
import com.mindfusion.spreadsheet.Measure;
import com.mindfusion.spreadsheet.ObjectAnchor;
import com.mindfusion.spreadsheet.ObjectAnchorPosition;
import com.mindfusion.spreadsheet.ObjectEvent;
import com.mindfusion.spreadsheet.ObjectInteractionEvent;
import com.mindfusion.spreadsheet.ObjectInteractionValidationEvent;
import com.mindfusion.spreadsheet.ObjectMouseEvent;
import com.mindfusion.spreadsheet.ObjectValidationEvent;
import com.mindfusion.spreadsheet.Picture;
import com.mindfusion.spreadsheet.RowMouseEvent;
import com.mindfusion.spreadsheet.Selection;
import com.mindfusion.spreadsheet.StyleData;
import com.mindfusion.spreadsheet.TrackChangesCommand;
import com.mindfusion.spreadsheet.ValidationType;
import com.mindfusion.spreadsheet.VerticalAlignment;
import com.mindfusion.spreadsheet.Workbook;
import com.mindfusion.spreadsheet.WorkbookAdapter;
import com.mindfusion.spreadsheet.WorkbookView;
import com.mindfusion.spreadsheet.WorkbookViewAdapter;
import com.mindfusion.spreadsheet.Worksheet;
import com.mindfusion.spreadsheet.WorksheetChangedEvent;
import com.mindfusion.spreadsheet.WorksheetEvent;
import com.mindfusion.spreadsheet.WorksheetMouseEvent;
import com.mindfusion.spreadsheet.standardforms.DialogResult;
import com.mindfusion.spreadsheet.standardforms.HyperlinkForm;


public class Features
{
	public static void main(String[] args)
	{
		try
		{
			UIManager.setLookAndFeel(UIManager.getSystemLookAndFeelClassName());
		}
		catch (Exception e)
		{
			e.printStackTrace();
		}

		SwingUtilities.invokeLater(new Runnable()
		{
			public void run()
			{
				MainWindow window = new MainWindow();
				window.setTitle("JSpreadsheet Sample: Features");
				window.setVisible(true);
			}
		});
	}
	
	public static class MainWindow extends JFrame implements ClipboardOwner
	{
		public MainWindow()
		{
			setDefaultCloseOperation(EXIT_ON_CLOSE);
			setSize(1280, 1024);

			JMenuBar menuBar = new JMenuBar();

			String[] features = new String[]
			{
				"Merging",
				"Freezing",
				"Grouping and Outlining",
				"Undo and redo",
				"Copy and paste",
				"Auto-filling",
				"Auto filter",
				"Annotations",
				"Hyperlinks",
				"Font customization",
				"Formatted text",
				"Data validation",
				"Number formats",
				"Styles",
				"Events",
				"Printing",
				"Calculations"
			};
			JMenu featuresMenu = new JMenu("Features");
			for (String feature: features)
			{
				JMenuItem menuItem = new JMenuItem(feature);
				menuItem.addActionListener(new ActionListener()
				{
					@Override
					public void actionPerformed(ActionEvent e)
					{
						onMenu(e.getSource());
					}
				});
				featuresMenu.add(menuItem);
			}
			menuBar.add(featuresMenu);

			setJMenuBar(menuBar);

			pane = new JDesktopPane();
			add(pane, BorderLayout.CENTER);

			// Merging
			mergingFrame = createMergingFrame();
			pane.add(mergingFrame);
			pane.getDesktopManager().maximizeFrame(mergingFrame);

			// Freezing
			freezingFrame = createFreezingFrame();
			pane.add(freezingFrame);
			pane.getDesktopManager().maximizeFrame(freezingFrame);

			// Outlining
			outlineFrame = createOutlineFrame();
			pane.add(outlineFrame);
			pane.getDesktopManager().maximizeFrame(outlineFrame);

			// Undo/redo
			undoRedoFrame = createUndoRedoFrame();
			pane.add(undoRedoFrame);
			pane.getDesktopManager().maximizeFrame(undoRedoFrame);

			// Copy/paste
			copyPasteFrame = createCopyPasteFrame();
			pane.add(copyPasteFrame);
			pane.getDesktopManager().maximizeFrame(copyPasteFrame);

			// Autofilling
			autofillingFrame = createAutofillingFrame();
			pane.add(autofillingFrame);
			pane.getDesktopManager().maximizeFrame(autofillingFrame);

			// Auto filter
			autoFilterFrame = createAutoFilterFrame();
			pane.add(autoFilterFrame);
			pane.getDesktopManager().maximizeFrame(autoFilterFrame);

			// Annotation
			annotationsFrame = createAnnotationsFrame();
			pane.add(annotationsFrame);
			pane.getDesktopManager().maximizeFrame(annotationsFrame);

			// Hyperlinks
			hyperlinksFrame = createHyperlinksFrame();
			pane.add(hyperlinksFrame);
			pane.getDesktopManager().maximizeFrame(hyperlinksFrame);

			// Font customization
			fontCustomizationFrame = createFontCustomizationFrame();
			pane.add(fontCustomizationFrame);
			pane.getDesktopManager().maximizeFrame(fontCustomizationFrame);

			// Formatted text
			formattedTextFrame = createFormattedTextFrame();
			pane.add(formattedTextFrame);
			pane.getDesktopManager().maximizeFrame(formattedTextFrame);

			// Data validation
			dataValidationFrame = createDataValidationFrame();
			pane.add(dataValidationFrame);
			pane.getDesktopManager().maximizeFrame(dataValidationFrame);

			// Number formats
			numberFormatsFrame = createNumberFormatsFrame();
			pane.add(numberFormatsFrame);
			pane.getDesktopManager().maximizeFrame(numberFormatsFrame);

			// Styles
			stylesFrame = createStylesFrame();
			pane.add(stylesFrame);
			pane.getDesktopManager().maximizeFrame(stylesFrame);

			// Events
			eventsFrame = createEventsFrame();
			pane.add(eventsFrame);
			pane.getDesktopManager().maximizeFrame(eventsFrame);
			
			// Styles
			printingFrame = createPrintingFrame();
			pane.add(printingFrame);
			pane.getDesktopManager().maximizeFrame(printingFrame);

			// Styles
			calculationsFrame = createCalculationsFrame();
			pane.add(calculationsFrame);
			pane.getDesktopManager().maximizeFrame(calculationsFrame);
			
			addWindowListener(new WindowAdapter() {
				@Override
				public void windowOpened(WindowEvent e)
				{
					super.windowOpened(e);
					showFrame("Merging");
				}
			});
		}

		@Override
		public void lostOwnership(Clipboard clipboard, Transferable contents)
		{
		}

		private JInternalFrame createMergingFrame()
		{
			Workbook workbook = new Workbook();
			Worksheet sheet = workbook.getWorksheets().add("Merging");
			
			JInternalFrame frame = new JInternalFrame("Merging", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();

			IStyle style = sheet.getCellRanges().get(0, 0, 0x400, 0x100000).getStyle();
			style.setFontName("Calibri");
			style.setFontBold(true);
			style.setFontSize(14.0);
			style.setHorizontalAlignment(HorizontalAlignment.Center);
			style.setVerticalAlignment(VerticalAlignment.Middle);

			SolidBrush mergeBrush = new SolidBrush(new Color(156, 170, 198));

			sheet.getCells().get("A1").setData("A1:D1");
			sheet.getCellRanges().get("A1:D1").merge();
			sheet.getCellRanges().get("A1:D1").getStyle().setBackground(mergeBrush);

			sheet.getCells().get("A3").setData("A3:A7");
			sheet.getCellRanges().get("A3:A7").merge();
			sheet.getCellRanges().get("A3:A7").getStyle().setBackground(mergeBrush);

			sheet.getCells().get("A9").setData("A9:C12");
			sheet.getCellRanges().get("A9:C12").merge();
			sheet.getCellRanges().get("A9:C12").getStyle().setBackground(mergeBrush);

			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			view.getSelection().set(0, 0, 3, 0);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Merge/Unmerge");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Worksheet activeSheet = view.getActiveWorksheet();
					CellRange selectedRange = view.getSelection().getRange();

					CellRange merge = activeSheet.getMergedCell(selectedRange.getLeft(), selectedRange.getTop());
					if (merge == null || merge.getLeft() != selectedRange.getLeft() || merge.getTop() != selectedRange.getTop() ||
						merge.getRight() != selectedRange.getRight() || merge.getBottom() != selectedRange.getBottom())
					{
						if (!activeSheet.merge(selectedRange))
							activeSheet.unmerge(selectedRange);
					}
					else
					{
						activeSheet.unmerge(selectedRange);
					}
				}
			});
			optionsPanel.add(button);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			return frame;
		}

		private JInternalFrame createFreezingFrame()
		{
			Workbook workbook = new Workbook();
			Worksheet sheet = workbook.getWorksheets().add("Freezing");

			JInternalFrame frame = new JInternalFrame("Freezing", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();

			sheet.getColumns().get(0, 6).setWidth(Measure.point(80));
			sheet.getRows().get(0).setHeight(Measure.point(20));
			sheet.getRows().get(202).setHeight(Measure.point(20));

			sheet.getCells().get("A1").setData("Order Number");
			sheet.getCells().get("B1").setData("Invoice Number");
			sheet.getCells().get("C1").setData("Unit Price");
			sheet.getCells().get("D1").setData("Quantity");
			sheet.getCells().get("E1").setData("Subtotal");
			sheet.getCells().get("F1").setData("VAT");
			sheet.getCells().get("G1").setData("Total");

			NumberFormat formatter = NumberFormat.getCurrencyInstance();
			for (int i = 0; i <= 200; i++)
			{ 
				String row = Integer.toString(i + 2);
				sheet.getCells().get("A" + row).setData("'" + String.format("%1$08d", i + 1));
				sheet.getCells().get("B" + row).setData(String.format("'201402%1$05d", i + 1));
				sheet.getCells().get("C" + row).setData(formatter.format(rnd.nextInt(27) + 2 + (rnd.nextInt(87) + 12) % 3 == 0 ? 0.49 : 0.99));
				sheet.getCells().get("D" + row).setData(rnd.nextInt(4) + 1);
				sheet.getCells().get("E" + row).setData(String.format("=C%1$s*D%1$s", row));
				sheet.getCells().get("F" + row).setData(String.format("=E%1$s*0.2", row));
				sheet.getCells().get("G" + row).setData(String.format("=E%1$s*1.2", row));
			}

			sheet.getCells().get("A203").setData("February Orders");
			sheet.getCells().get("G203").setData("=SUM(G2:G202)");

			sheet.getCellRanges().get("A203:F203").merge();

			sheet.setFrozenRowCount(1);

			IStyle style = sheet.getRows().get(0).getStyle();
			style.setFontBold(true);
			style.setFontSize(10.0);
			style.setHorizontalAlignment(HorizontalAlignment.Center);
			style.setVerticalAlignment(VerticalAlignment.Middle);

			style = sheet.getRows().get(202).getStyle();
			style.setFontBold(true);
			style.setFontItalic(true);
			style.setFontSize(10.0);
			style.setHorizontalAlignment(HorizontalAlignment.Right);
			style.setVerticalAlignment(VerticalAlignment.Middle);
			style.setPaddingRight(Measure.point(5));

			style = sheet.getCellRanges().get(0, 0, 6, 202).getStyle();
			style.setBorderBottomColor(SystemColor.activeCaptionBorder);
			style.setBorderLeftColor(SystemColor.activeCaptionBorder);
			style.setBorderRightColor(SystemColor.activeCaptionBorder);
			style.setBorderTopColor(SystemColor.activeCaptionBorder);

			style = sheet.getCellRanges().get(0, 0, 6, 0).getStyle();
			style.setBorderBottomSize(Measure.point(2));
			style.setBorderLeftSize(Measure.point(2));
			style.setBorderRightSize(Measure.point(2));
			style.setBorderTopSize(Measure.point(2));
			style = sheet.getCellRanges().get(0, 202, 6, 202).getStyle();
			style.setBorderBottomSize(Measure.point(2));
			style.setBorderLeftSize(Measure.point(2));
			style.setBorderRightSize(Measure.point(2));
			style.setBorderTopSize(Measure.point(2));

			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			view.getSelection().set(0, 0, 3, 0);
			view.setScrollY(190);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Freeze/Unfreeze");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Worksheet activeSheet = view.getActiveWorksheet();
					Cell activeCell = view.getActiveCell();
					if (activeSheet.getFrozenColumnCount() > 0 || activeSheet.getFrozenRowCount() > 0)
					{
						activeSheet.setFrozenColumnCount(0);
						activeSheet.setFrozenRowCount(0);
					}
					else
					{
						if (activeCell.getColumn() > 20 || activeCell.getRow() > 20)
						{
							JOptionPane.showMessageDialog(MainWindow.this, "The maximum number of frozen columns and rows is 20.");
						}
						else
						{
							activeSheet.setFrozenColumnCount(activeCell.getColumn());
							activeSheet.setFrozenRowCount(activeCell.getRow());
						}
					}
				}
			});
			optionsPanel.add(button);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			return frame;
		}

		private JInternalFrame createOutlineFrame()
		{
			Workbook workbook = new Workbook();

			JInternalFrame frame = new JInternalFrame("Grouping and Outlining", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			InputStream stream = MainWindow.class.getResourceAsStream("Outlines.xml");
			if (stream != null)
			{
				try
				{
					DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
				    Document document = builder.parse(stream);
					workbook.loadFromXml(document);
					stream.close();
				}
				catch (ParserConfigurationException | SAXException | IOException e)
				{
				}
			}

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Group rows");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Selection selection = view.getSelection();
					TrackChangesCommand command = workbook.startChangeOperation(view.getActiveWorksheet());
					try
					{
						if (!view.getActiveWorksheet().getRows().get(selection.getTop(), selection.getBottom()).group())
							JOptionPane.showMessageDialog(MainWindow.this, "Unable to group the selected rows.");
					}
					finally
					{
						if (command != null)
							command.dispose();
					}
				}
			});
			optionsPanel.add(button);
			button = new JButton("Ungroup rows");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Selection selection = view.getSelection();
					TrackChangesCommand command = workbook.startChangeOperation(view.getActiveWorksheet());
					try
					{
						if (!view.getActiveWorksheet().getRows().get(selection.getTop(), selection.getBottom()).ungroup())
							JOptionPane.showMessageDialog(MainWindow.this, "Unable to ungroup the selected rows.");
					}
					finally
					{
						if (command != null)
							command.dispose();
					}
				}
			});
			optionsPanel.add(button);
			button = new JButton("Group columns");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Selection selection = view.getSelection();
					TrackChangesCommand command = workbook.startChangeOperation(view.getActiveWorksheet());
					try
					{
						if (!view.getActiveWorksheet().getColumns().get(selection.getLeft(), selection.getRight()).group())
							JOptionPane.showMessageDialog(MainWindow.this, "Unable to group the selected columns.");
					}
					finally
					{
						if (command != null)
							command.dispose();
					}
				}
			});
			optionsPanel.add(button);
			button = new JButton("Ungroup columns");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Selection selection = view.getSelection();
					TrackChangesCommand command = workbook.startChangeOperation(view.getActiveWorksheet());
					try
					{
						if (!view.getActiveWorksheet().getColumns().get(selection.getLeft(), selection.getRight()).ungroup())
							JOptionPane.showMessageDialog(MainWindow.this, "Unable to ungroup the selected columns.");
					}
					finally
					{
						if (command != null)
							command.dispose();
					}
				}
			});
			optionsPanel.add(button);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			return frame;
		}

		private JInternalFrame createUndoRedoFrame()
		{
			final Workbook workbook = new Workbook();
			workbook.setUndoEnabled(true);
			Worksheet sheet = workbook.getWorksheets().add("Undo/Redo");

			JInternalFrame frame = new JInternalFrame("Undo and Redo", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			InputStream stream = MainWindow.class.getResourceAsStream("UndoRedo.csv");
			if (stream != null)
			{
				CsvImporter importer = new CsvImporter();
				try
				{
					importer.setSeparator('\t');
					importer.setQuote('"');
					importer.importData(stream, sheet);
				}
				catch (IOException e1)
				{
					e1.printStackTrace();
				}
			}

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			final JButton undoButton = new JButton("Undo");
			undoButton.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					workbook.undo();
				}
			});
			undoButton.setEnabled(false);
			optionsPanel.add(undoButton);
			final JButton redoButton = new JButton("Redo");
			redoButton.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					workbook.redo();
				}
			});
			redoButton.setEnabled(false);
			optionsPanel.add(redoButton);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			workbook.addPropertyChangedListener(new PropertyChangeListener()
			{
				@Override
				public void propertyChange(PropertyChangeEvent e)
				{
					if (Objects.equals(e.getPropertyName(), "CanUndo"))
						undoButton.setEnabled((boolean)e.getNewValue());
					if (Objects.equals(e.getPropertyName(), "CanRedo"))
						redoButton.setEnabled((boolean)e.getNewValue());
				}
			});

			return frame;
		}

		private JInternalFrame createCopyPasteFrame()
		{
			final Workbook workbook = new Workbook();
			workbook.setUndoEnabled(true);
			final Worksheet sheet = workbook.getWorksheets().add("Copy/Paste");

			JInternalFrame frame = new JInternalFrame("Copy and Paste", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();
			IStyle headingStyle = sheet.getCellRanges().get("A1:E1").getStyle();
			headingStyle.setBackground(new SolidBrush(Colors.DarkSlateBlue));
			headingStyle.setTextColor(Colors.White);
			headingStyle.setFontBold(true);
			IStyle contentStyle = sheet.getCellRanges().get("A2:E10").getStyle();
			contentStyle.setFormat("0.##");
			contentStyle.setBackground(new SolidBrush(Colors.LightSteelBlue));
			for (int i = 0; i < 5; i++)
			{
				sheet.getCells().get(i, 0).setData(String.format("Heading%1$d", i + 1));
				for (int j = 1; j < 10; j++)
					sheet.getCells().get(i, j).setData(rnd.nextDouble());
			}
			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Copy");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					CellRange selectionRange = view.getSelection().getRange();

					// Copy data
					CellData cellData = sheet.copyData(selectionRange);

					// Copy styles
					StyleData styleData = sheet.copyStyles(selectionRange);

					Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
					try
					{
						String data = cellData.serialize() + "<>" + styleData.serialize();
						clipboard.setContents(new StringSelection(data), MainWindow.this);
					}
					catch (ParserConfigurationException | TransformerException ex)
					{
						ex.printStackTrace();
					}
				}
			});
			optionsPanel.add(button);
			button = new JButton("Paste");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Clipboard clipboard = Toolkit.getDefaultToolkit().getSystemClipboard();
					Transferable data = clipboard.getContents(this);
					if (data.isDataFlavorSupported(DataFlavor.stringFlavor))
					{
						try
						{
							String contents = (String)data.getTransferData(DataFlavor.stringFlavor);
							String[] parts = contents.split("<>");

							CellData cellData = new CellData();
							cellData.deserialize(parts[0]);

							StyleData styleData = new StyleData();
							styleData.deserialize(parts[1]);

							sheet.pasteData(view.getActiveCell(), cellData);
							sheet.pasteStyles(view.getActiveCell(), styleData);
						}
						catch (Exception ex)
						{
						}
					}
				}
			});
			optionsPanel.add(button);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			return frame;
		}

		private JInternalFrame createAutofillingFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Auto-filling");

			JInternalFrame frame = new JInternalFrame("Auto-filling", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();
			sheet.getCells().get(0, 0).setData(1);
			sheet.getCells().get(1, 0).setData(2);
			sheet.getCells().get(0, 1).setData(4);
			sheet.getCells().get(1, 1).setData(6);
			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			view.getSelection().set(0, 0, 1, 1);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			return frame;
		}

		private JInternalFrame createAutoFilterFrame()
		{
			final Workbook workbook = new Workbook();

			JInternalFrame frame = new JInternalFrame("Auto-filter", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			InputStream stream = MainWindow.class.getResourceAsStream("AutoFilter.xml");
			if (stream != null)
			{
				try
				{
					DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
				    Document document = builder.parse(stream);
					workbook.loadFromXml(document);
					stream.close();
				}
				catch (ParserConfigurationException | SAXException | IOException e)
				{
				}
			}

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			return frame;
		}

		private JInternalFrame createAnnotationsFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Annotations");

			JInternalFrame frame = new JInternalFrame("Annotations", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();

			Cell cell = sheet.getCells().get(3, 3);
			cell.setData("Styles");
			cell.setAnnotation("Annotation with applied styles.");
			CellAnnotation annotation = cell.getAnnotation().getCellAnnotation();
			annotation.setShowAlways(true);
			annotation.setAnchorPosition(ObjectAnchorPosition.TopLeft);
			annotation.setHorizontalOffset(Measure.point(-75));
			annotation.setVerticalOffset(Measure.point(-30));
			annotation.setHeight(Measure.point(50));
			annotation.setBackColor(new Color (45, 57, 86));
			annotation.setTextColor(Colors.White);
			annotation.setLineColor(new Color (90, 121, 165));
			annotation.setMargin(Measure.point(5));
			annotation.setFontBold(true);

			cell = sheet.getCells().get(4, 3);
			cell.setData("Default");
			cell.setAnnotation("Annotation without customizations.");
			annotation = cell.getAnnotation().getCellAnnotation();

			cell = sheet.getCells().get(4, 4);
			cell.setData("Fading and animations");
			cell.setAnnotation("This annotation uses animation and fading.");
			annotation = cell.getAnnotation().getCellAnnotation();
			annotation.setAnchorPosition(ObjectAnchorPosition.BottomRight);
			annotation.setVerticalOffset(Measure.point(5));
			annotation.setIndicatorColor(Colors.DarkGreen);

			cell = sheet.getCells().get(3, 4);
			cell.setData("More styles");
			cell.setAnnotation("C:\\> echo Hello, world!\r\nHello, world!\r\nC:\\>\u005F");
			annotation = cell.getAnnotation().getCellAnnotation();
			annotation.setShowAlways(true);
			annotation.setAnchorPosition(ObjectAnchorPosition.BottomLeft);
			annotation.setHorizontalOffset(Measure.point(-140));
			annotation.setVerticalOffset(Measure.point(5));
			annotation.setWidth(Measure.point(135));
			annotation.setHeight(Measure.point(40));
			annotation.setLineStyle(DashStyle.Dash);
			annotation.setLineSize(Measure.point(1.5));
			annotation.setLineColor(Colors.Chartreuse);
			annotation.setFontName("Lucida Console");
			annotation.setTextColor(Colors.Chartreuse);
			annotation.setBackColor(Colors.Black);
			annotation.setFontSize(12.0);
			annotation.setFontBold(true);

			sheet.getCellRanges().get(4, 4, 5, 4).merge();

			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			view.getSelection().addObject(annotation);

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Add/remove annotation");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Cell cell = view.getActiveCell();
					if (cell == null)
						return;

					if (cell.getAnnotation() == null)
					{
						// Add a new annotation
						cell.setAnnotation("New comment.");
						cell.getAnnotation().setShowAlways(true);
						view.beginEdit(cell.getAnnotation());
					}
					else
					{
						// Remove the annotation
						cell.removeAnnotation();
					}
				}
			});
			optionsPanel.add(button);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			return frame;
		}

		private JInternalFrame createHyperlinksFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Hyperlinks");

			JInternalFrame frame = new JInternalFrame("Hyperlinks", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			Cell cell = sheet.getCells().get(1, 1);
			cell.setData("Created with Cell.Hyperlink");
			cell.setHyperlink(HyperlinkType.WebAddress, "http://www.mindfusion.eu", "");

			cell = sheet.getCells().get(1, 2);
			cell.setData("=HYPERLINK(\"http://mindfusion.eu/Forum/YaBB.pl\", \"Created with the HYPERLINK function\")");

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Add/remove hyperlink");
			JButton button2 = new JButton("Edit hyperlink");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Cell cell = view.getActiveCell();
					if (cell == null)
						return;

					if (cell.getHyperlink() == null)
					{
						HyperlinkForm hyperlinkForm = new HyperlinkForm(MainWindow.this,
							workbook, cell.getValue() != null ? cell.getValue().toString() : "");
						hyperlinkForm.setVisible(true);
						if (hyperlinkForm.getDialogResult() == DialogResult.OK)
						{
							cell.setHyperlink(hyperlinkForm.getHyperlinkType(), hyperlinkForm.getTarget(), hyperlinkForm.getSubTarget());
							cell.getHyperlink().setToolTip(hyperlinkForm.getToolTip());
							button2.setEnabled(true);
						}
					}
					else
					{
						cell.removeHyperlink();
						button2.setEnabled(false);
					}
				}
			});
			optionsPanel.add(button);
			button2.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					Cell cell = view.getActiveCell();
					if (cell == null || cell.getHyperlink() == null)
						return;

					HyperlinkForm hyperlinkForm = new HyperlinkForm(MainWindow.this, workbook, cell.getHyperlink());
					hyperlinkForm.setVisible(true);
					if (hyperlinkForm.getDialogResult() == DialogResult.OK)
					{
						cell.getHyperlink().setType(hyperlinkForm.getHyperlinkType());
						cell.getHyperlink().setTarget(hyperlinkForm.getTarget());
						cell.getHyperlink().setSubTarget(hyperlinkForm.getSubTarget());
						cell.getHyperlink().setToolTip(hyperlinkForm.getToolTip());
					}
				}
			});
			optionsPanel.add(button2);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			view.addViewListener(new WorkbookViewAdapter()
			{
				@Override
				public void hyperlinkClicked(CellEvent e)
				{
					if (e.getCell().getHyperlink().getType() != HyperlinkType.CurrentWorkbook)
					{
						try
						{
							Desktop desktop = Desktop.isDesktopSupported() ? Desktop.getDesktop() : null;
						    if (desktop != null && desktop.isSupported(Desktop.Action.BROWSE))
					            desktop.browse(new URI(e.getCell().getHyperlink().getTarget()));
						}
						catch (Exception ex)
						{
							JOptionPane.showMessageDialog(MainWindow.this, ex.getMessage());
						}
					}
				}

				@Override
				public void activeCellChanged(EventObject e)
				{
					button2.setEnabled(view.getActiveCell().getHyperlink() != null);
				}
			});

			return frame;
		}

		private JInternalFrame createFontCustomizationFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Font customization");

			JInternalFrame frame = new JInternalFrame("Font customization", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();
			sheet.getRows().get(1).getStyle().setFontBold(true);
			sheet.getCells().get(1, 1).setData("Font");
			sheet.getCells().get(2, 1).setData("Result");

			for (int i = 0; i < 9; i++)
			{
				Cell cell = sheet.getCells().get(2, i + 2);
				cell.setData(LOREM_IPSUM[i]);
				generateRandomFont(cell);
				sheet.getCells().get(1, i + 2).setData(getFontInfo(cell));
			}
			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			view.resizeColumnsToFit(1, 2);
			view.resizeRowsToFit(2, 11);

			return frame;
		}

		private static final String[] LOREM_IPSUM = new String[] 
		{
			"Lorem ipsum dolor sit amet, consectetur adipiscing elit.",
			"Morbi ultrices sollicitudin lectus, sit amet tincidunt dui aliquet vel.",
			"Vestibulum accumsan quam volutpat pharetra porta.",
			"Donec in quam a mi eleifend tincidunt rutrum nec velit.",
			"Proin sit amet tortor felis.",
			"Morbi et nisi vel felis rutrum porta.",
			"Vivamus bibendum, orci eu vestibulum imperdiet, est mauris pulvinar libero, id tincidunt felis ligula nec massa.",
			"Ut a cursus leo. Duis molestie elit mi.",
			"Nullam volutpat dictum mi, vel aliquet augue eleifend ut.",
			"Donec rhoncus arcu vel nunc venenatis, eget semper neque placerat."
		};

		private void generateRandomFont(Cell cell)
		{
			if (fonts == null)
				fonts = GraphicsEnvironment.getLocalGraphicsEnvironment().getAvailableFontFamilyNames();

			IStyle style = cell.getStyle();

			style.setFontName(fonts[rnd.nextInt(fonts.length)]);
			style.setFontBold(rnd.nextInt(2) % 2 == 0);
			style.setFontItalic(rnd.nextInt(2) % 2 == 0);
			style.setFontSize(Double.valueOf(rnd.nextInt(8) + 10));
		}

		private String[] fonts;

		private String getFontInfo(Cell cell)
		{
			IStyle style = cell.getStyle();

			return String.format("%1$s %2$.0fpt", style.getFontName(), style.getFontSize());
		}

		private JInternalFrame createFormattedTextFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Formatted text");

			JInternalFrame frame = new JInternalFrame("Formatted Text", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.getCells().get(0, 0).setData(new FormattedText(
				"The formatted texts can contain <b>bold</b>, <i>italic</i>, <u>underlined</u> and " +
				"<s>strikeout</s> texts as well as <color=\"#ff0000\">text</color> <color=\"#00ff00\">in</color> " +
				"<color=\"#0000ff\">various</color> <color=\"#ff8000\">colors</color>, " +
				"<fontname=\"Courier New\">fonts</fontname> and <fontsize=\"16\">sizes</fontsize>." + System.lineSeparator() +
				"<color=\"#8000FF\"><s>Or</s> <fontname=\"Times New Roman\"><fontsize=\"14\">any</fontsize> " +
				"<b>comb<i>inat</i>ion</b></color> <fontsize=\"18\">of</fontname> <color=\"#008080\">the</color></fontsize> a<u>bov</u>e."));

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			view.resizeColumnToFit(0);
			view.resizeRowToFit(0);

			return frame;
		}

		private JInternalFrame createDataValidationFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Data validation");

			JInternalFrame frame = new JInternalFrame("Data Validation", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);
			
			Cell cell = sheet.getCells().get(0, 0);
			cell.setData("Select a value from the list in A2");

			sheet.getCellRanges().get(0, 0, 2, 0).merge();

			cell = sheet.getCells().get(0, 1);
			cell.setData(1);
			IValidation validation = cell.getValidation();
			validation.setType(ValidationType.List);
			validation.setFirst("1,2,3,4");
			validation.setAllowBlankCells(false);
			validation.setShowDropdownList(true);
			validation.setShowInputMessage(true);
			validation.setInputMessage("Select a value from the list");
			validation.setShowError(true);
			validation.setErrorTitle("Invalid input.");
			validation.setErrorMessage("Enter a value from the specified list.");

			cell = sheet.getCells().get(1, 1);
			cell.setData("=IF(A2=1,\"one\",IF(A2=2,\"two\",IF(A2=3,\"three\",\"four\")))");
			cell.getStyle().setHorizontalAlignment(HorizontalAlignment.Right);

			sheet.getCellRanges().get(1, 1, 2, 1).merge();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			return frame;
		}

		private JInternalFrame createNumberFormatsFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Number formats");

			JInternalFrame frame = new JInternalFrame("Number Formats", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			sheet.beginInit();

			sheet.getRows().get(1).getStyle().setFontBold(true);

			sheet.getCells().get(1, 1).setData("Plain Value");
			sheet.getCells().get(2, 1).setData("Format");
			sheet.getCells().get(3, 1).setData("Formatted Value");

			DefaultNumericFormats defaultFormats = workbook.getDefaultNumericFormats();
			sheet.getCells().get(2, 2).setData("'" + defaultFormats.getPercentFormats().get(1).getFormat());
			sheet.getCells().get(2, 3).setData("'" + defaultFormats.getNumberFormats().get(7).getFormat());
			sheet.getCells().get(2, 4).setData("'" + defaultFormats.getNumberFormats().get(4).getFormat());
			// TODO: this is null atm
//			sheet.getCells().get(2, 5).setData("'" + defaultFormats.getCurrencyFormats().get(1).getFormat());
			sheet.getCells().get(2, 6).setData("'" + defaultFormats.getDateFormats().get(0).getFormat());
			sheet.getCells().get(2, 7).setData("'" + defaultFormats.getDateFormats().get(2).getFormat());
			sheet.getCells().get(2, 8).setData("'" + defaultFormats.getScientificFormats().get(0).getFormat());
			sheet.getCells().get(2, 9).setData("'" + defaultFormats.getFractionFormats().get(1).getFormat());
			sheet.getCells().get(2, 10).setData("???.???");
			sheet.getCells().get(2, 11).setData("Boolean");

			for (int r = 2; r <= 11; r++)
			{
				sheet.getCells().get(1, r).setData(Math.pow(r, 2) * r * 0.1);
				sheet.getCells().get(3, r).setData("=B" + Integer.toString(r + 1));
				sheet.getCells().get(3, r).getStyle().setFormat(sheet.getCells().get(2, r).getValue() != null ?
					StringUtilities.trimStart(sheet.getCells().get(2, r).getValue().toString(), '\'')  : "General");
			}

			sheet.endInit();

			sheet.getColumns().get(2, 3).setWidth(Measure.point(150));
			sheet.getColumns().get(2).getStyle().setHorizontalAlignment(HorizontalAlignment.Left);

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			workbook.addWorkbookListener(new WorkbookAdapter()
			{
				@Override
				public void worksheetCellChanged(CellChangedEvent e)
				{
					if (e.getCell().getColumn() == 2 && e.getCell().getRow() >= 2 && e.getCell().getRow() <= 11)
						sheet.getCells().get(3, e.getCell().getRow()).getStyle().setFormat(e.getCell().getValue() != null ? e.getCell().getValue().toString() : "General");
				}

				@Override
				public void worksheetCellsCleared(WorksheetEvent e)
				{
					for (int r = 2; r <= 11; r++)
					{
						sheet.getCells().get(3, r).getStyle().setFormat(e.getWorksheet().getCells().get(2, r).getValue() != null ?
							e.getWorksheet().getCells().get(2, r).getValue().toString() : "General");
					}
				}
			});

			return frame;
		}

		private JInternalFrame createStylesFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Styles");

			JInternalFrame frame = new JInternalFrame("Styles", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			IStyle sheetStyle = sheet.getCellRanges().get(0, 0, 0x400, 0x100000).getStyle();
			sheetStyle.setFontName("Calibri");
			sheet.getCellRanges().get(4, 4, 0x400, 0x100000).getStyle().setBackground(new SolidBrush(Colors.WhiteSmoke));

			IStyle rangeStyle = sheet.getCellRanges().get(1, 1, 5, 5).getStyle();
			rangeStyle.setFontSize(15.0);
			rangeStyle.setTextColor(new Color(0x30, 0x30, 0x30));
			rangeStyle.setBackground(new SolidBrush(new Color ( 225, 225, 225)));
			rangeStyle.setBorderBottomColor(new Color (206, 0, 0));
			rangeStyle.setBorderLeftColor(new Color (206, 0, 0));
			rangeStyle.setBorderRightColor(new Color (206, 0, 0));
			rangeStyle.setBorderTopColor(new Color (206, 0, 0));

			IStyle rowStyle = sheet.getRows().get(0, 3).getStyle();
			rowStyle.setBackground(new SolidBrush(new Color (0, 52, 102)));
			rowStyle.setFontUnderline(true);
			rowStyle.setFontSize(18.0);
			rowStyle.setTextColor(Colors.White);
			rowStyle.setHorizontalAlignment(HorizontalAlignment.Center);
			rowStyle.setVerticalAlignment(VerticalAlignment.Middle);
			rowStyle.setBorderBottomColor(new Color (224, 233, 233));
			rowStyle.setBorderLeftColor(new Color (224, 233, 233));
			rowStyle.setBorderRightColor(new Color (224, 233, 233));
			rowStyle.setBorderTopColor(new Color (224, 233, 233));

			IStyle columnStyle = sheet.getColumns().get(0, 3).getStyle();
			columnStyle.setBackground(new SolidBrush(new Color (224, 233, 233)));
			columnStyle.setTextColor(Colors.Gray);
			columnStyle.setFontItalic(true);
			columnStyle.setFontSize(18.0);
			columnStyle.setHorizontalAlignment(HorizontalAlignment.Right);
			columnStyle.setVerticalAlignment(VerticalAlignment.Middle);
			columnStyle.setBorderBottomColor(new Color(90, 121, 165));
			columnStyle.setBorderLeftColor(new Color(90, 121, 165));
			columnStyle.setBorderRightColor(new Color(90, 121, 165));
			columnStyle.setBorderTopColor(new Color(90, 121, 165));

			sheet.getColumns().get(0).setWidth(Measure.point(55));
			sheet.getColumns().get(1, 5).setWidth(Measure.point(70));
			sheet.getRows().get(0).setHeight(Measure.point(25));
			sheet.getRows().get(1, 5).setHeight(Measure.point(18));

			sheet.beginInit();

			for (int c = 1; c <= 5; )
				sheet.getCells().get(c, 0).setData(String.format("Column %1$d", c++));

			for (int r = 1; r <= 5; )
				sheet.getCells().get(0, r).setData(String.format("Row %1$d", r++));

			for (int c = 1; c <= 5; c++)
			{
				for (int r = 1; r <= 5; r++)
					sheet.getCells().get(c, r).setData(String.format("=ADDRESS(%1$d,%2$d,4)", r + 1, c + 1));
			}

			sheet.endInit();

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			return frame;
		}

		private JInternalFrame createEventsFrame()
		{
			final Workbook workbook = new Workbook();
			final Worksheet sheet = workbook.getWorksheets().add("Events");

			JInternalFrame frame = new JInternalFrame("Events", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			Cell cell = sheet.getCells().get("A1");
			cell.setData("Click Me!");
			cell.setHyperlink(HyperlinkType.WebAddress, "", "");

			Picture pic = sheet.getDrawing().addPicture();
			try
			{
				pic.setSource(ImageIO.read(MainWindow.class.getResourceAsStream("SpreadsheetIcon.png")));
			}
			catch (IOException ex)
			{
			}
			pic.setAnchor(ObjectAnchor.ToCell);
			pic.setFromColumn(1);
			pic.setFromRow(0);
			pic.setHorizontalOffset(Measure.point(10));
			pic.setVerticalOffset(Measure.point(10));
			pic.setWidth(Measure.point(100));
			pic.setHeight(Measure.point(100));

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel panel = new JPanel();
			JList<String> eventList = new JList<>();
			final DefaultListModel<String> model = new DefaultListModel<>();
			eventList.setModel(model);
			eventList.setVisibleRowCount(20);
			JScrollPane scrollPane = new JScrollPane(eventList);
			final JScrollBar scrollbar = scrollPane.getVerticalScrollBar();
			panel.add(scrollPane);
			frame.getContentPane().add(panel, BorderLayout.LINE_END);

			workbook.addWorkbookListener(new WorkbookAdapter()
			{
				@Override
				public void worksheetCellChanged(CellChangedEvent e)
				{
					model.addElement("worksheetCellChanged");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void worksheetCellChanging(CellValidationEvent e)
				{
					model.addElement("worksheetCellChanging");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void worksheetCellsCleared(WorksheetEvent e)
				{
					model.addElement("worksheetCellsCleared");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void worksheetChanged(WorksheetChangedEvent e)
				{
					model.addElement("worksheetChanged");
					scrollbar.setValue(scrollbar.getMaximum());
				}
			});

			view.addViewListener(new WorkbookViewAdapter()
			{
				@Override
				public void activeCellChanged(EventObject e)
				{
					model.addElement("activeCellChanged");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void activeWorksheetChanged(EventObject e)
				{
					model.addElement("activeWorksheetChanged");
					scrollbar.setValue(scrollbar.getMaximum());
				}
			
				@Override
				public void cellClicked(CellMouseEvent e)
				{
					model.addElement("cellClicked");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void columnClicked(ColumnMouseEvent e)
				{
					model.addElement("columnClicked");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void hyperlinkClicked(CellEvent e)
				{
					model.addElement("hyperlinkClicked");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void inplaceEditEnded(InplaceEditEvent e)
				{
					model.addElement("inplaceEditEnded");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void inplaceEditEnding(InplaceEditValidationEvent e)
				{
					model.addElement("inplaceEditEnding");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void inplaceEditStarted(InplaceEditEvent e)
				{
					model.addElement("inplaceEditStarted");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void inplaceEditStarting(InplaceEditValidationEvent e)
				{
					model.addElement("inplaceEditStarting");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectClicked(ObjectMouseEvent e)
				{
					model.addElement("objectClicked");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectDeleted(ObjectEvent e)
				{
					model.addElement("objectDeleted");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectDeleting(ObjectValidationEvent e)
				{
					model.addElement("objectDeleting");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectDeselected(ObjectEvent e)
				{
					model.addElement("objectDeselected");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectModified(ObjectInteractionEvent e)
				{
					model.addElement("objectModified");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectModifying(ObjectInteractionValidationEvent e)
				{
					model.addElement("objectModifying");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectSelected(ObjectEvent e)
				{
					model.addElement("objectSelected");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectSelecting(ObjectInteractionValidationEvent e)
				{
					model.addElement("objectSelecting");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void objectStartModifying(ObjectInteractionValidationEvent e)
				{
					model.addElement("objectStartModifying");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void rowClicked(RowMouseEvent e)
				{
					model.addElement("rowClicked");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void selectionChanged(EventObject e)
				{
					model.addElement("selectionChanged");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void worksheetAdded(WorksheetEvent e)
				{
					model.addElement("worksheetAdded");
					scrollbar.setValue(scrollbar.getMaximum());
				}

				@Override
				public void worksheetTabClicked(WorksheetMouseEvent e)
				{
					model.addElement("worksheetTabClicked");
					scrollbar.setValue(scrollbar.getMaximum());
				}
			});

			return frame;
		}

		private JInternalFrame createPrintingFrame()
		{
			final Workbook workbook = new Workbook();

			JInternalFrame frame = new JInternalFrame("Printing", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			InputStream stream = MainWindow.class.getResourceAsStream("Printing.xml");
			if (stream != null)
			{
				try
				{
					DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
				    Document document = builder.parse(stream);
					workbook.loadFromXml(document);
					stream.close();
				}
				catch (ParserConfigurationException | SAXException | IOException e)
				{
				}
			}

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			JPanel optionsPanel = new JPanel();
			JButton button = new JButton("Print");
			button.addActionListener(new ActionListener()
			{
				@Override
				public void actionPerformed(ActionEvent e)
				{
					workbook.print();
				}
			});
			optionsPanel.add(button);
			frame.getContentPane().add(optionsPanel, BorderLayout.PAGE_END);

			return frame;
		}

		private JInternalFrame createCalculationsFrame()
		{
			final Workbook workbook = new Workbook();

			JInternalFrame frame = new JInternalFrame("Calculations", true, true, true, true);
			frame.setDefaultCloseOperation(WindowConstants.HIDE_ON_CLOSE);

			InputStream stream = MainWindow.class.getResourceAsStream("Calculations.xml");
			if (stream != null)
			{
				try
				{
					DocumentBuilder builder = DocumentBuilderFactory.newInstance().newDocumentBuilder();
				    Document document = builder.parse(stream);
					workbook.loadFromXml(document);
					stream.close();
				}
				catch (ParserConfigurationException | SAXException | IOException e)
				{
				}
			}

			final WorkbookView view = new WorkbookView();
			view.setWorkbook(workbook);
			frame.getContentPane().setLayout(new BorderLayout());
			frame.getContentPane().add(view, BorderLayout.CENTER);
			frame.pack();

			return frame;
		}

		private void onMenu(Object source)
		{
			JMenuItem item = (JMenuItem)source;
			showFrame(item.getText());
		}

		private void showFrame(String name)
		{
			if (lastFrame != null)
				lastFrame.setVisible(false);

			JInternalFrame nextFrame = null;
			switch (name)
			{
			case "Merging":
				nextFrame = mergingFrame;
				break;
			case "Freezing":
				nextFrame = freezingFrame;
				break;
			case "Grouping and Outlining":
				nextFrame = outlineFrame;
				break;
			case "Undo and redo":
				nextFrame = undoRedoFrame;
				break;
			case "Copy and paste":
				nextFrame = copyPasteFrame;
				break;
			case "Auto-filling":
				nextFrame = autofillingFrame;
				break;
			case "Auto filter":
				nextFrame = autoFilterFrame;
				break;
			case "Annotations":
				nextFrame = annotationsFrame;
				break;
			case "Hyperlinks":
				nextFrame = hyperlinksFrame;
				break;
			case "Font customization":
				nextFrame = fontCustomizationFrame;
				break;
			case "Formatted text":
				nextFrame = formattedTextFrame;
				break;
			case "Data validation":
				nextFrame = dataValidationFrame;
				break;
			case "Number formats":
				nextFrame = numberFormatsFrame;
				break;
			case "Styles":
				nextFrame = stylesFrame;
				break;
			case "Events":
				nextFrame = eventsFrame;
				break;
			case "Printing":
				nextFrame = printingFrame;
				break;
			case "Calculations":
				nextFrame = calculationsFrame;
				break;
			default:
				return;
			}

			if (nextFrame != null)
			{
				pane.getDesktopManager().maximizeFrame(nextFrame);
				nextFrame.setVisible(true);
				lastFrame = nextFrame;
			}
		}


		private static final long serialVersionUID = 1L;

		private final Random rnd = new Random(DateTime.now().getMillisecond());

		private JDesktopPane pane;
		private JInternalFrame lastFrame;
		private JInternalFrame mergingFrame;
		private JInternalFrame freezingFrame;
		private JInternalFrame outlineFrame;
		private JInternalFrame undoRedoFrame;
		private JInternalFrame copyPasteFrame;
		private JInternalFrame autofillingFrame;
		private JInternalFrame autoFilterFrame;
		private JInternalFrame annotationsFrame;
		private JInternalFrame hyperlinksFrame;
		private JInternalFrame fontCustomizationFrame;
		private JInternalFrame formattedTextFrame;
		private JInternalFrame dataValidationFrame;
		private JInternalFrame numberFormatsFrame;
		private JInternalFrame stylesFrame;
		private JInternalFrame eventsFrame;
		private JInternalFrame printingFrame;
		private JInternalFrame calculationsFrame;
	}
}